import greenfoot.*;
import java.awt.Color;
import java.awt.image.BufferedImage;

/**
 * An actor which displays the mandelbrot set.
 * 
 * @author Sven van Nigtevecht
 * @version 1.0
 */
public class MandelbrotView extends Actor
{
    private static final double R_MAX = 2.0;
    private static final double R_MAX_SQUARED = R_MAX *R_MAX;
    private static final int T_MAX = 1000;
    
    private int maxColor;
    private volatile boolean calculating;
    
    private ComplexNumber c;
    private ComplexNumber z;
    
    private volatile double scale;
    private double extraX,extraY;
    
    private BufferedImage image;
    private int drawingX, drawingY;
    private int width, height;
    private int halfWidth, halfHeight;
    
    /**
     * Create a new MandelbrotView.
     * @param width The width of the view.
     * @param height The height of the view.
     */
    public MandelbrotView(int width, int height)
    {
        GreenfootImage img = getImage();
        img.clear();
        this.width = width;
        this.height = height;
        img.scale(width, height);
        this.halfWidth = width /2;
        this.halfHeight = height /2;
        
        this.calculating = true;
        
        this.c = new ComplexNumber();
        this.z = new ComplexNumber();
        
        this.scale = 100.0;
        this.extraX = 0.0;
        this.extraY = 0.0;
        
        final MandelbrotView self = this;
        final ScrollListener sl = new ScrollListener();
        sl.setRunnable(new Runnable() {
            public void run() {
                self.scale -= (double)sl.getRotationAmount() *self.scale/10.0;
                if (self.scale < 20.0)
                    self.scale = 20.0;
                self.resetDrawing();
            }
        });
        sl.start();
    }
    
    /**
     * Render a new pixel and listen to key presses.
     */
    public void act()
    {
        if (drawingX >= width) {
            drawingX = 0;
            drawingY++;
        }
        if (drawingY == height) {
            drawingX = 0;
            drawingY = 0;
            calculating = !calculating;
        }
        
        configureC(drawingX, drawingY);
        z.normal = 0.0;
        z.i = 0.0;
        
        int time;
        for (time = 1; time < T_MAX; time++) {
            z.multiply(z);
            z.add(c);
            double dist = z.absSquared();
            if (dist > R_MAX_SQUARED) {
                if (maxColor < time && calculating) {
                    maxColor = time;
                }
                break;
            }
        }
        if (!calculating) {
            if (time >= T_MAX) {
                time = 0;
            }
            image = getImage().getAwtImage();
            image.setRGB(drawingX, drawingY, convertToColor(time,z));
        }
        drawingX++;
        
        listenToKeys();
    }
    
    private void configureC(int pixelX, int pixelY)
    {
        c.normal = (double)(pixelX -halfWidth) /scale +this.extraX;
        c.i = (double)(pixelY -halfHeight) /scale +this.extraY;
    }
    
    private int convertToColor(int time, ComplexNumber zn)
    {
        if (time == 0) {
            return (255 << 24);
        } else {
            double dist = zn.abs();
            double color = time +1 -Math.log(Math.log(dist)) /Math.log(2) /T_MAX;
            return Color.HSBtoRGB(0.5f +12.0f *(float)color, 9.5f, 0.9f);
        }
    }
    
    private void resetDrawing()
    {
        drawingX = 0;
        drawingY = 0;
        calculating = true;
    }
    
    private void listenToKeys()
    {
        String key = Greenfoot.getKey();
        if ("left".equals(key)) {
            extraX -= 5.0/scale;
        }
        if ("right".equals(key)) {
            extraX += 5.0/scale;
        }
        if ("down".equals(key)) {
            extraY += 5.0/scale;
        }
        if ("up".equals(key)) {
            extraY -= 5.0/scale;
        }
    }
}